import asyncio
import pickle
from numbers import Number
from typing import Any, List, Optional, Text, Tuple

from simple_pymq.broker.base import Broker
from simple_pymq.config import logger
from simple_pymq.exceptions import FullError, EmptyError


is_redis_installed = True
try:
    from redis.asyncio import BlockingConnectionPool, Redis
except ImportError:
    is_redis_installed = False


class RedisBroker(Broker):
    def __init__(
        self,
        name: Text = "RedisBroker",
        maxsize: int = 0,
        *args,
        host: Optional[Text] = None,
        port: Optional[int] = None,
        db: Optional[int] = None,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        connection_pool: Optional["BlockingConnectionPool"] = None,
        key_base_name: Optional[Text] = None,
        key_prefix: Text = "",
        key_postfix: Text = "",
        key_expire: int = 60 * 60 * 24 * 7,
        block: bool = True,
        timeout: Optional[Number] = None,
        **kwargs,
    ):
        if is_redis_installed is False:
            raise ImportError("Package 'aiofiles' is not installed.")

        super(RedisBroker, self).__init__(
            name=name, maxsize=maxsize, *args, block=block, timeout=timeout, **kwargs
        )

        self.host = host
        self.port = port
        self.db = db
        self.redis_conn_pool = connection_pool
        self.key_name = (
            key_prefix + (key_base_name if key_base_name else self.name) + key_postfix
        )
        self.key_expire = key_expire

        if self.redis_conn_pool is not None:
            self.redis_conn_pool = connection_pool
        elif self.host is not None:
            self.redis_conn_pool = BlockingConnectionPool(
                host=self.host,
                port=self.port,
                db=self.db,
                max_connections=20,
                username=username,
                password=password,
            )
        else:
            raise ValueError("Either 'connection_pool' or 'host' must be provided.")

        self.redis_client = Redis(connection_pool=self.redis_conn_pool)
        logger.debug(
            f"{self.name}: {self.redis_conn_pool.connection_kwargs.get('username')}"
            + f":******@{self.redis_conn_pool.connection_kwargs.get('host')}"
            + f":{self.redis_conn_pool.connection_kwargs.get('port')}"
            + f".{self.redis_conn_pool.connection_kwargs.get('db')}"
        )

    async def qsize(self) -> int:
        return await self.redis_client.llen(self.key_name)

    async def empty(self) -> bool:
        count = await self.qsize()
        return True if count == 0 else False

    async def full(self) -> bool:
        if self.maxsize <= 0:
            return False

        count = await self.qsize()
        return True if count >= self.maxsize else False

    async def get_nowait(self) -> Any:
        data: List[bytes] = await self.redis_client.rpop(self.key_name, count=1)

        if data is None or len(data) == 0:
            raise EmptyError("Queue is empty.")

        assert (len(data) == 1, "Only one item should be returned.")
        value_bytes = data[0]

        return pickle.loads(value_bytes)

    async def get(
        self, block: Optional[bool] = None, timeout: Optional[Number] = None
    ) -> Any:
        block = self.block if block is None else block
        timeout = self.timeout if timeout is None else timeout

        if block is True:
            data: Optional[Tuple[bytes, bytes]] = await self.redis_client.brpop(
                self.key_name, timeout=timeout
            )

            if data is None or len(data) == 0:
                raise asyncio.TimeoutError("Queue is empty.")

            _, value_bytes = data
            item = pickle.loads(value_bytes)
        else:
            item = await self.get_nowait()

        return item

    async def put_nowait(self, item: Any) -> None:
        if await self.full() is True:
            raise FullError("Queue is full.")
        pipe = self.redis_client.pipeline()
        pipe.lpush(self.key_name, pickle.dumps(item))
        pipe.expire(self.key_name, time=self.key_expire)
        await pipe.execute()

    async def put(
        self, item: Any, block: Optional[bool] = None, timeout: Optional[Number] = None
    ) -> None:
        block = self.block if block is None else block
        timeout = self.timeout if timeout is None else timeout

        async def _wait_put(item: Any):
            while True:
                if await self.full() is True:
                    await asyncio.sleep(0.05)
                else:
                    pipe = self.redis_client.pipeline()
                    pipe.lpush(self.key_name, pickle.dumps(item))
                    pipe.expire(self.key_name, time=self.key_expire)
                    await pipe.execute()
                    break

        if block is True:
            try:
                await asyncio.wait_for(_wait_put(item=item), timeout=timeout)
            except asyncio.TimeoutError:
                raise asyncio.TimeoutError("Queue is full.")

        else:
            item = await self.put_nowait(item=item)

    async def join(self) -> None:
        while True:
            if self.empty() is True:
                break
            else:
                asyncio.sleep(0.05)

    async def task_done(self) -> None:
        pass



import logging
import math
from abc import ABC
from typing import Any, Optional, Text, Type

from simple_pymq.broker.base import Broker
from simple_pymq.config import settings


logger = logging.getLogger(settings.logger_name)


class Consumer(ABC):
    def __init__(
        self,
        name: Text = "Consumer",
        *args,
        block: bool = True,
        timeout: Optional[float] = None,
        max_consume_count: int = 0,
        **kwargs,
    ):
        self.name = name or self.__class__.__name__
        self.max_consume_count = 0 if max_consume_count <= 0 else int(max_consume_count)
        self.block = block
        self.timeout = timeout

    async def listen(
        self,
        broker: Type[Broker],
        *args,
        block: Optional[bool] = None,
        timeout: Optional[float] = None,
        max_consume_count: Optional[int] = None,
        raise_error: bool = True,
        **kwargs,
    ):
        block = self.block if block is None else block
        timeout = self.timeout if timeout is None else timeout
        max_consume_count = (
            self.max_consume_count if max_consume_count is None else max_consume_count
        )
        max_consume_count = (
            float("inf")
            if max_consume_count <= 0 or math.isinf(max_consume_count)
            else math.ceil(max_consume_count)
        )

        consume_count = 0
        while True:
            item = await broker.get(block=block, timeout=timeout)

            try:
                await self.consume(item=item, broker=broker)

            except Exception as e:
                logger.exception(e)
                logger.error(f"Consumer '{self.name}' raise error: {str(e)}")
                if raise_error is True:
                    raise e

            finally:
                consume_count += 1

            if consume_count >= max_consume_count:
                break

    async def consume(self, item: Any):
        raise NotImplementedError


class NullConsumer(Consumer):
    def __init__(
        self,
        name: Text = "NullConsumer",
        *args,
        block: bool = True,
        timeout: Optional[float] = None,
        max_consume_count: int = 0,
        **kwargs,
    ):
        super(NullConsumer, self).__init__(
            name=name,
            *args,
            block=block,
            timeout=timeout,
            max_consume_count=max_consume_count,
            **kwargs,
        )

    async def consume(self, item: Any, broker: Type[Broker]):
        await broker.task_done()


class PrintConsumer(Consumer):
    def __init__(
        self,
        name: Text = "PrintConsumer",
        *args,
        block: bool = True,
        timeout: Optional[float] = None,
        max_consume_count: int = 0,
        **kwargs,
    ):
        super(PrintConsumer, self).__init__(
            name=name,
            *args,
            block=block,
            timeout=timeout,
            max_consume_count=max_consume_count,
            **kwargs,
        )

    async def consume(self, item: Any, broker: Type[Broker]):
        print(item)
        logger.info(item)
        await broker.task_done()



import asyncio
from abc import ABC
from numbers import Number
from typing import Any, Optional, Text

from simple_pymq.exceptions import FullError, EmptyError


class Broker(ABC):
    def __init__(
        self,
        name: Text = "Broker",
        maxsize: int = 0,
        *args,
        block: bool = True,
        timeout: Optional[Number] = None,
        **kwargs
    ):
        self.name = name
        self.maxsize = maxsize
        self.block = block
        self.timeout = timeout

    async def empty(self) -> bool:
        raise NotImplementedError

    async def full(self) -> bool:
        raise NotImplementedError

    async def get(self, block: bool = True, timeout: Optional[Number] = None) -> Any:
        raise NotImplementedError

    async def get_nowait(self) -> Any:
        raise NotImplementedError

    async def join(self) -> None:
        raise NotImplementedError

    async def put(
        self, item: Any, block: bool = True, timeout: Optional[Number] = None
    ) -> None:
        raise NotImplementedError

    async def put_nowait(self, item: Any) -> None:
        raise NotImplementedError

    async def qsize(self) -> int:
        raise NotImplementedError

    async def task_done(self) -> None:
        raise NotImplementedError


class QueueBroker(Broker):
    def __init__(
        self,
        name: Text = "QueueBroker",
        maxsize: int = 0,
        block: bool = True,
        timeout: Optional[Number] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        *args,
        **kwargs
    ):
        super(QueueBroker, self).__init__(
            name=name, maxsize=maxsize, *args, block=block, timeout=timeout, **kwargs
        )
        self.name = name
        self.maxsize = maxsize
        self.queue = asyncio.Queue(maxsize=maxsize, loop=loop)

    async def empty(self) -> bool:
        return self.queue.empty()

    async def full(self) -> bool:
        return self.queue.full()

    async def get(
        self, block: Optional[bool] = None, timeout: Optional[Number] = None
    ) -> Any:
        block = self.block if block is None else block
        timeout = self.timeout if timeout is None else timeout

        if block is True:
            item = await asyncio.wait_for(self.queue.get(), timeout=timeout)
        else:
            item = self.queue.get_nowait()

        return item

    async def get_nowait(self) -> Any:
        try:
            item = self.queue.get_nowait()
            return item
        except asyncio.QueueEmpty:
            raise EmptyError

    async def join(self) -> None:
        await self.queue.join()

    async def put(
        self, item: Any, block: Optional[bool] = None, timeout: Optional[Number] = None
    ) -> None:
        block = self.block if block is None else block
        timeout = self.timeout if timeout is None else timeout

        if block is True:
            await asyncio.wait_for(self.queue.put(item), timeout=timeout)
        else:
            try:
                self.queue.put_nowait(item)
            except asyncio.QueueFull:
                raise FullError

    async def put_nowait(self, item: Any):
        try:
            self.queue.put_nowait(item)
        except asyncio.QueueFull:
            raise FullError

    async def qsize(self) -> int:
        return self.queue.qsize()

    async def task_done(self) -> None:
        self.queue.task_done()



import asyncio
import logging
from abc import ABC
from numbers import Number
from typing import Any, Optional, Text, Type

from simple_pymq.broker.base import Broker
from simple_pymq.config import settings
from simple_pymq.exceptions import FullError


logger = logging.getLogger(settings.logger_name)


class Producer(ABC):
    def __init__(
        self,
        name: Text = "Producer",
        *args,
        block: bool = True,
        timeout: Optional[Number] = None,
        **kwargs,
    ):
        self.name = name
        self.timeout = timeout
        self.block = block

    async def produce(
        self,
        broker: Type[Broker],
        *args,
        block: Optional[bool] = None,
        timeout: Optional[Number] = None,
        **kwargs,
    ):
        raise NotImplementedError


class TimeCounterProducer(Producer):
    def __init__(
        self,
        name: Text = "TimeCounterProducer",
        block: bool = True,
        timeout: Optional[Number] = None,
        count_seconds: float = 1.0,
        max_produce_count: Optional[int] = None,
        put_value: Any = 1,
        *args,
        **kwargs,
    ):
        super(TimeCounterProducer, self).__init__(
            name=name, *args, block=block, timeout=timeout, **kwargs
        )
        self.count_seconds = count_seconds
        self.max_produce_count = max_produce_count
        self.put_value = put_value

    async def produce(
        self,
        broker: Type[Broker],
        *args,
        block: Optional[bool] = None,
        timeout: Optional[Number] = None,
        count_seconds: Optional[float] = None,
        max_produce_count: Optional[int] = None,
        put_value: Any = None,
        ignore_full_error: bool = False,
        raise_full_error: bool = False,
        **kwargs,
    ):
        timeout = self.timeout if timeout is None else timeout
        block = self.block if block is None else block
        count_seconds = self.count_seconds if count_seconds is None else count_seconds
        max_produce_count = max_produce_count or self.max_produce_count or float("inf")
        put_value = self.put_value if put_value is None else put_value

        count = 0
        while True:
            try:
                await broker.put(put_value, block=block, timeout=timeout)

            except FullError as e:
                if raise_full_error is True:
                    raise e
                if ignore_full_error is True:
                    logger.info(
                        f"The broker '{broker}' is full, "
                        + f"so skip the item '{put_value}'."
                    )
                else:
                    logger.error(
                        f"The broker '{broker}' is full, "
                        + f"so skip the item '{put_value}'."
                    )

            finally:
                count += 1

            if count >= max_produce_count:
                break

            await asyncio.sleep(count_seconds)



import asyncio
import pickle
import time
from numbers import Number
from pathlib import Path
from simple_pymq.broker.base import Broker
from typing import Any, AsyncGenerator, List, Optional, Text, Type

from simple_pymq.exceptions import FullError, EmptyError


is_aiofiles_installed = True
try:
    import aiofiles
except ImportError:
    is_aiofiles_installed = False


class SimpleFileLock:
    def __init__(
        self,
        lock_path: Path,
        block: bool = True,
        timeout: Optional[Number] = None,
        poll_interval: float = 0.05,
        lock_timeout: float = 5.0,
    ):
        self.lock_path = Path(lock_path).resolve()
        self.block = block
        self.timeout = timeout
        self.poll_interval = poll_interval
        self.lock_timeout = lock_timeout

    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.release()

    async def acquire(
        self,
        block: Optional[bool] = None,
        timeout: Optional[Number] = None,
        poll_interval: Optional[float] = None,
    ) -> bool:
        block = self.block if block is None else block
        timeout = self.timeout if timeout is None else timeout
        poll_interval = self.poll_interval if poll_interval is None else poll_interval

        if self.lock_path.parent.exists() is False:
            self.lock_path.parent.mkdir(parents=True)

        lock_acquired = await asyncio.wait_for(
            self._touch_lock(block=block, poll_interval=poll_interval), timeout=timeout
        )
        return lock_acquired

    async def release(self):
        self.lock_path.unlink(missing_ok=True)

    async def locked(self):
        is_locked = True

        if self.lock_path.exists() is True:
            if time.time() - self.lock_path.stat().st_mtime >= self.lock_timeout:
                is_locked = False
            else:
                is_locked = True
        else:
            is_locked = False

        return is_locked

    async def _touch_lock(self, block: bool, poll_interval: float) -> bool:
        while True:
            try:
                self.lock_path.touch(exist_ok=False)
                return True

            except FileExistsError:
                if time.time() - self.lock_path.stat().st_mtime >= self.lock_timeout:
                    self.lock_path.touch(exist_ok=True)
                    return True
                elif block is False:
                    return False

                await asyncio.sleep(poll_interval)


class SimpleFileBroker(Broker):
    def __init__(
        self,
        file: Text,
        name: Text = "SimpleFileBroker",
        maxsize: int = 0,
        *args,
        block: bool = True,
        timeout: Optional[Number] = None,
        lock_timeout: float = 5.0,
        **kwargs
    ):
        if is_aiofiles_installed is False:
            raise ImportError("Package 'aiofiles' is not installed.")

        super(SimpleFileBroker, self).__init__(
            name=name, maxsize=maxsize, *args, block=block, timeout=timeout, **kwargs
        )

        self.file = Path(file).resolve().absolute()
        self.lock_path = self.file.with_name(self.file.name + ".lock")
        self.lock_timeout = lock_timeout
        self.unfinished = 0

        if self.file.parent.exists() is False:
            self.file.parent.mkdir(parents=True)
        if self.file.exists() is False:
            with open(self.file, "wb") as f:
                f.write(pickle.dumps([]))

    async def empty(self) -> bool:
        count = await self._lock_count()
        return True if count == 0 else False

    async def full(self) -> bool:
        count = await self._lock_count()
        return True if count >= self.maxsize else False

    async def get(
        self, block: Optional[bool] = None, timeout: Optional[Number] = None
    ) -> Any:
        block = self.block if block is None else block
        timeout = self.timeout if timeout is None else timeout

        if block is True:
            item = await asyncio.wait_for(self._lock_wait_and_get(), timeout=timeout)
        else:
            item = await self.get_nowait()

        return item

    async def get_nowait(self) -> Any:
        return await self._lock_get_nowait()

    async def join(self) -> None:
        while True:
            if self.empty() is True:
                break
            else:
                asyncio.sleep(0.05)

    async def put(
        self, item: Any, block: Optional[bool] = None, timeout: Optional[Number] = None
    ) -> None:
        block = self.block if block is None else block
        timeout = self.timeout if timeout is None else timeout

        if block is True:
            item = await asyncio.wait_for(
                self._lock_wait_and_put(item=item), timeout=timeout
            )
        else:
            item = await self.put_nowait(item=item)

    async def put_nowait(self, item: Any) -> None:
        await self._lock_put_nowait(item=item)

    async def qsize(self) -> int:
        return await self._lock_count()

    async def task_done(self) -> None:
        self.unfinished -= 1

    async def _lock_wait_and_get(self) -> Any:
        while True:
            async with SimpleFileLock(
                self.lock_path, block=True, lock_timeout=self.lock_timeout
            ):
                items = await self._read_pickle()

                if len(items) > 0:
                    item = items.pop(0)
                    await self._dump_pickle(items=items)
                    return item

            await asyncio.sleep(0.05)

    async def _lock_get_nowait(self) -> Any:
        async with SimpleFileLock(
            self.lock_path, block=False, lock_timeout=self.lock_timeout
        ):
            items = await self._read_pickle()
            if len(items) > 0:
                item = items.pop(0)
                await self._dump_pickle(items=items)
                return item
            else:
                raise EmptyError

    async def _lock_wait_and_put(self, item: Any) -> None:
        while True:
            async with SimpleFileLock(
                self.lock_path, block=True, lock_timeout=self.lock_timeout
            ):
                items = await self._read_pickle()
                if len(items) < self.maxsize:
                    items.append(item)
                    await self._dump_pickle(items=items)
                    return None

            await asyncio.sleep(0.05)

    async def _lock_put_nowait(self, item: Any) -> None:
        async with SimpleFileLock(
            self.lock_path, block=False, lock_timeout=self.lock_timeout
        ):
            items = await self._read_pickle()
            if len(items) < self.maxsize:
                items.append(item)
                await self._dump_pickle(items=items)
                return None
            else:
                raise FullError

    async def _lock_count(self) -> int:
        async with SimpleFileLock(
            self.lock_path, block=True, lock_timeout=self.lock_timeout
        ):
            items = await self._read_pickle()
            return len(items)

    async def _read_pickle(self) -> List[Any]:
        async with aiofiles.open(self.file, "rb") as f:
            data = await f.read()
            return pickle.loads(data)

    async def _dump_pickle(self, items: List[Any]) -> None:
        async with aiofiles.open(self.file, "wb") as f:
            await f.write(pickle.dumps(items))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['simple_pymq',
 'simple_pymq.broker',
 'simple_pymq.consumer',
 'simple_pymq.exceptions',
 'simple_pymq.pipeline',
 'simple_pymq.producer',
 'simple_pymq.utils']

package_data = \
{'': ['*']}

install_requires = \
['pytz', 'rich']

extras_require = \
{':extra == "all"': ['redis[redis]>=4.2'],
 ':extra == "all" or extra == "cron"': ['croniter[cron]'],
 ':extra == "all" or extra == "fs"': ['aiofiles[fs]']}

setup_kwargs = {
    'name': 'simple-pymq',
    'version': '0.3.0',
    'description': 'Simple python message queue framework is ready to serve.',
    'long_description': '# simple-pymq #\n\n[![dockhardman](https://circleci.com/gh/dockhardman/simple-pymq.svg?style=shield)](https://app.circleci.com/pipelines/github/dockhardman/simple-pymq)\n\nSimple python message queue framework is ready to serve.\n\n## Installation ##\n\n```bash\npip install simple-pymq\n```\n\n## Usage ##\n\nSimple message queue pipeline in memory:\n\n```python\nimport asyncio\nfrom simple_pymq import (\n    PrintConsumer,\n    QueueBroker,\n    SimpleMessageQueue,\n    TimeCounterProducer,\n)\n\n\nasync def main():\n    q = QueueBroker(maxsize=32)\n    p = TimeCounterProducer(\n        count_seconds=1.0, max_produce_count=3, put_value="Message here."\n    )\n    c = PrintConsumer(max_consume_count=3)\n    mq = SimpleMessageQueue()\n\n    await mq.run(broker=q, producers=p, consumers=c)\n    print("All tasks done!")\n\n\nasyncio.run(main())\n```\n',
    'author': 'Allen Chou',
    'author_email': 'f1470891079@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8.0,<3.11.0',
}


setup(**setup_kwargs)



import os
import logging
import uuid
from pathlib import Path
from typing import Any, Optional, Text, Union

from rich.console import Console


def get_os_env(
    key: Text, *args, default: Any, return_type: Text = "str"
) -> Optional[Union[Text, bool, int, float, Path, uuid.UUID]]:
    return_type = return_type.casefold()
    value = os.environ.get(key)

    if value is None:
        valid_value = default
    elif return_type == "str" or return_type == "text":
        valid_value = value
    elif return_type == "int":
        valid_value = int(value)
    elif return_type == "float":
        valid_value = float(value)
    elif return_type == "path":
        valid_value = Path(value)
    elif return_type == "bool":
        valid_value = bool(value)
    elif return_type == "uuid":
        valid_value = uuid.UUID(value)
    else:
        raise ValueError(f"Unknown return type: '{return_type}'")

    return valid_value


class Settings:
    # Common
    logger_name: Text = get_os_env("logger_name", default="simple-pymq")


settings = Settings()
logger = logging.getLogger(settings.logger_name)
console = Console()



import asyncio
from abc import ABC
from typing import List, Type, Union

from simple_pymq.broker.base import Broker
from simple_pymq.consumer.base import Consumer
from simple_pymq.producer.base import Producer
from simple_pymq.config import logger


class MessageQueue(ABC):
    def __init__(self, *args, **kwargs):
        pass

    async def run(
        self,
        broker: Type[Broker],
        *args,
        producers: Union[List[Type[Producer]], Producer],
        consumers: Union[List[Type[Consumer]], Consumer],
        **kwargs,
    ):
        raise NotImplementedError


class SimpleMessageQueue(MessageQueue):
    def __init__(self, *args, **kwargs):
        pass

    async def run(
        self,
        broker: Type[Broker],
        *args,
        producers: Union[List[Type[Producer]], Producer],
        consumers: Union[List[Type[Consumer]], Consumer],
        **kwargs,
    ):
        producers: List[Type[Producer]] = (
            [producers] if isinstance(producers, Producer) else producers
        )
        consumers: List[Type[Consumer]] = (
            [consumers] if isinstance(consumers, Consumer) else consumers
        )

        if len(producers) == 0:
            raise ValueError(f"Should provide at least one producer.")
        if len(consumers) == 0:
            raise ValueError(f"Should provide at least one consumer.")

        tasks: List["asyncio.Task"] = []

        for producer in producers:
            _producer_task = asyncio.create_task(
                producer.produce(broker=broker, *args, **kwargs)
            )
            tasks.append(_producer_task)

        for consumer in consumers:
            _consumer_task = asyncio.create_task(
                consumer.listen(broker=broker, *args, **kwargs)
            )
            tasks.append(_consumer_task)

        task_results = await asyncio.gather(*tasks, return_exceptions=True)

        for task_result in task_results:
            if isinstance(task_result, Exception):
                logger.exception(task_result)
